from __future__ import absolute_import


import stregion
from os.path import join

try:
    from astropy.io import fits as pyfits
except ImportError:
    from astropy.io import fits as pyfits

import numpy as np

from stregion.wcs_helper import fix_lon

# At some point, pyfits.Card.fromstring has changed from unbound
# method to bounded method.

if pyfits.Card.fromstring.__self__: # 
    def pyfits_card_fromstring(l):
        return pyfits.Card.fromstring(l)
else:
    def pyfits_card_fromstring(l):
        c = pyfits.Card()
        return c.fromstring(l)

rootdir = "examples"

def demo_header():
    cards = pyfits.CardList()
    for l in open(join(rootdir, "sample_fits01.header")):
        card = pyfits_card_fromstring(l.strip())
        cards.append(card)
    h = pyfits.Header(cards)
    return h

def test_region():

    ref_region_name = "test01_img.reg"

    region_list = ["test01_fk5_sexagecimal.reg",
                   "test01_gal.reg",
                   "test01_ds9_physical.reg",
                   "test01_fk5_degree.reg",
                   "test01_mixed.reg",
                   "test01_ciao.reg",
                   "test01_ciao_physical.reg",
                   ]

    header = demo_header()

    ref_region = stregion.open(join(rootdir,ref_region_name)).as_imagecoord(header)

    for reg_name in region_list:
        r = stregion.open(join(rootdir,reg_name)).as_imagecoord(header)
        for reg0, reg in zip(ref_region, r):
            if reg.name == "rotbox":
                reg.name = "box"

            assert reg0.name == reg.name
            if reg0.name in ["ellipse", "box"]:
                assert np.allclose(reg0.coord_list[:-1], reg.coord_list[:-1],
                                   atol=0.01)
                a0 = reg0.coord_list[-1]
                a1 = fix_lon(reg.coord_list[-1], 0)
                assert np.allclose([a0], [a1], atol=0.02)
            else:
                assert np.allclose(reg0.coord_list, reg.coord_list,
                                   atol=0.01)
            assert reg0.exclude == reg.exclude
        



from __future__ import print_function
from .version import __version__
from .ds9_region_parser import RegionParser
from .wcs_converter import check_wcs as _check_wcs
from itertools import cycle

_builtin_open = open

class ShapeList(list):
    def __init__(self, *ka, **kw):
        self._comment_list = kw.pop("comment_list", None)
        list.__init__(self, *ka, **kw)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return ShapeList(list.__getitem__(self, key))
        else:
            return list.__getitem__(self, key)

    def __getslice__(self, i, j):
        return self[max(0, i):max(0, j):]

    def check_imagecoord(self):
        if [s for s in self if s.coord_format != "image"]:
            return False
        else:
            return True

    def as_imagecoord(self, header, rot_wrt_axis = 1):
        """
        Return a new ShapeList where the coordinate of the each shape
        is converted to the image coordinate using the given header
        information
        """

        comment_list = self._comment_list
        if comment_list is None:
            comment_list = cycle([None])

        r = RegionParser.sky_to_image(list(zip(self, comment_list)),
                                      header, rot_wrt_axis = rot_wrt_axis)
        shape_list, comment_list = list(zip(*list(r)))
        return ShapeList(shape_list, comment_list=comment_list)


    def get_mpl_patches_texts(self, properties_func=None,
                              text_offset=5.0,
                              origin=1):
        """
        Often, the regions files implicitly assume the lower-left
        corner of the image as a coordinate (1,1). However, the python
        convetion is that the array index starts from 0. By default
        (origin = 1), coordinates of the returned mpl artists have
        coordinate shifted by (1, 1). If you do not want this shift,
        use origin=0.
        """
        from .mpl_helper import as_mpl_artists
        patches, txts = as_mpl_artists(self, properties_func,
                                       text_offset,
                                       origin=origin)

        return patches, txts

    def get_filter(self, header=None, origin=1, rot_wrt_axis=1):
        """
        Often, the regions files implicitly assume the lower-left
        corner of the image as a coordinate (1,1). However, the python
        convetion is that the array index starts from 0. By default
        (origin = 1), coordinates of the returned mpl artists have
        coordinate shifted by (1, 1). If you do not want this shift,
        use origin=0.
        """

        from .region_to_filter import as_region_filter

        if header is None:
            if not self.check_imagecoord():
                raise RuntimeError("the region has non-image coordinate. header is required.")
            reg_in_imagecoord = self
        else:
            reg_in_imagecoord = self.as_imagecoord(header, rot_wrt_axis = rot_wrt_axis)

        region_filter = as_region_filter(reg_in_imagecoord, origin=1)

        return region_filter


    def get_mask(self, hdu=None, header=None, shape=None, rot_wrt_axis=1):
        """
        creates a 2-d mask.

        get_mask(hdu=f[0])
        get_mask(shape=(10,10))
        get_mask(header=f[0].header, shape=(10,10))
        """

        if hdu and header is None:
            header = hdu.header
        if hdu and shape is None:
            shape = hdu.data.shape

        region_filter = self.get_filter(header=header, rot_wrt_axis=rot_wrt_axis)
        mask = region_filter.mask(shape)

        return mask
#
#    def write(self,outfile):
#        """ Writes the current shape list out as a region file """
#
#        # check for consistent coordinate system
#        if len(set([shape.coord_format for shape in self])) > 1:
#            raise ValueError("Inconsistent coordinate formats")
#
#        outf = _builtin_open(outfile,'w')
#
#        attr0 = self[0].attr[1]
#        defaultline = " ".join(["%s=%s" % (a,attr0[a]) for a in attr0 if a!='text'])
#
#        # first line is globals
#        print >>outf,"global",defaultline
#        # second line must be a coordinate format
#        print >>outf,self[0].coord_format
#
#        for shape in self:
#            text_coordlist = ["%f" % f for f in shape.coord_list]
#            if shape.comment:
#                print >>outf, shape.name + "(" + ",".join(text_coordlist) + \
#                              ") # " + shape.comment
#            else:
#                print >>outf, shape.name + "(" + ",".join(text_coordlist) + ")"
#
#        outf.close()


    def write(self,outfile):
        """ Writes the current shape list out as a region file """
        if len(self) < 1:
            print("WARNING: The region list is empty. The region file \"%s\" "
                           "will be empty." % outfile)
            try:
                outf = open(outfile,'w')
                outf.close()
                return
            except IOError as e:
                cmsg = "Unable to create region file \'%s\'." % outfile
                if e.args:
                    e.args = (e.args[0] + "\n" + cmsg,) + e.args[1:]
                else:
                    e.args=(cmsg,)
                raise e
            except:
                raise

        prev_cs = self[0].coord_format

        outf = None
        try:
            outf = _builtin_open(outfile,'w')

            attr0 = self[0].attr[1]
            defaultline = " ".join( [ "%s=%s" % (a,attr0[a]) for a in attr0 \
                                     if a!='text' ] )

            # first line is globals
            print("global", defaultline, file=outf)
            # second line must be a coordinate format
            print(prev_cs, file=outf)

            for shape in self:
                shape_attr = '' if prev_cs == shape.coord_format \
                    else shape.coord_format+"; "
                shape_excl = '-' if shape.exclude else ''
                text_coordlist = [ "%f" % f for f in shape.coord_list ]
                shape_coords = "(" + ",".join(text_coordlist) + ")"
                shape_comment = " # " + shape.comment if shape.comment else ''

                shape_str = shape_attr + shape_excl + shape.name + \
                            shape_coords + shape_comment

                print >>outf, shape_str

        except IOError as e:
            cmsg = "Unable to create region file \'%s\'." % outfile
            if e.args:
                e.args = (e.args[0] + "\n" + cmsg,) + e.args[1:]
            else:
                e.args=(cmsg,)
            if outf: outf.close()
            raise e
        except:
            if outf: outf.close()
            raise

        outf.close()

def parse(region_string):
    """
    Parse the input string of a ds9 region definition.
    Returns a list of Shape instances.
    """
    rp = RegionParser()
    ss = rp.parse(region_string)
    sss1 = rp.convert_attr(ss)
    sss2 = _check_wcs(sss1)

    shape_list, comment_list = rp.filter_shape2(sss2)
    return ShapeList(shape_list, comment_list=comment_list)

def open(fname):
    region_string = _builtin_open(fname).read()
    return parse(region_string)


# def parse_deprecated(region_string):
#     rp = RegionParser()
#     return rp.parseString(region_string)


def read_region(s):
    rp = RegionParser()
    ss = rp.parse(s)
    sss1 = rp.convert_attr(ss)
    sss2 = _check_wcs(sss1)

    return rp.filter_shape(sss2)


def read_region_as_imagecoord(s, header, rot_wrt_axis=1):
    rp = RegionParser()
    ss = rp.parse(s)
    sss1 = rp.convert_attr(ss)
    sss2 = _check_wcs(sss1)
    sss3 = rp.sky_to_image(sss2, header, rot_wrt_axis = rot_wrt_axis)

    return rp.filter_shape(sss3)


def get_mask(region, hdu, origin=1):
    """
    f = pyfits.read("test.fits")
    reg = read_region_as_imagecoord(s, f[0].header)
    mask = get_mask(reg, f[0])
    """

    from stregion.region_to_filter import as_region_filter

    data = hdu.data
    #header = hdu.header

    region_filter = as_region_filter(region, origin=origin)

    mask = region_filter.mask(data)

    return mask


if __name__ == '__main__':
    #reg = stregion.open("../mos_fov.reg")
    import stregion
    proposed_fov = 'fk5;circle(290.96388,14.019167,843.31194")'
    reg = stregion.parse(proposed_fov)
    reg_imagecoord = reg.as_imagecoord(header)
    patches, txts = reg.get_mpl_patches_texts()
    m = reg.get_mask(hdu=f[0])



import matplotlib.pyplot as plt
import matplotlib.cm as cm

import stregion
try:
    from astropy.io import fits as pyfits
except ImportError:
    from astropy.io import fits as pyfits

# read in the image
xray_name="pspc_skyview.fits"
f_xray = pyfits.open(xray_name)

try:
    import pywcsgrid2
    ax=pywcsgrid2.subplot(111, header=f_xray[0].header)
except ImportError:
    ax=plt.subplot(111)

ax.imshow(f_xray[0].data,
          cmap=cm.gray, vmin=0., vmax=0.00038, origin="lower")

reg_name = "test.reg"
r = stregion.open(reg_name).as_imagecoord(f_xray[0].header)

patch_list, text_list = r.get_mpl_patches_texts()
for p in patch_list:
    ax.add_patch(p)
for t in text_list:
    ax.add_artist(t)

plt.show()





import pyparsing

from pyparsing import Literal, CaselessKeyword, CaselessLiteral, \
     Word, Optional, OneOrMore, Group, Combine, ZeroOrMore, nums, \
     Forward, StringEnd, restOfLine, alphas, alphanums, Empty, \
     Or


def _unsigned_simple_number():
    # fnumber : 102.43, 12304.3e10,
    #           .32???
    point = Literal( "." )
    e     = Literal("e") | Literal("E") #CaselessLiteral( "E" )
    fnumber = Combine( Word(nums) +
                       Optional( point + Optional( Word( nums ) ) ) +
                       Optional( e + Word( "+-"+nums, nums ) ) )

    return fnumber #.setParseAction(lambda s,l,t: (float(t[0]), t[0]))
    #return fnumber.leaveWhitespace().setParseAction(lambda s,l,t: [ float(t[0])])



#simple_number = _simple_number()

usn = _unsigned_simple_number()

def getSign(s, l, tok):
    if not tok: return (1, "")
    if tok[0] == "-":
        return -1, "-"
    else:
        return 1, "+"

def get_default(v):
    def _f(s, l, tok):
        if tok:
            return tok
        return v
    return _f

optional_sign = Optional(Literal("+") | Literal("-")).setParseAction(get_default(""))
#optional_sign = Optional(Literal("+") | Literal("-")) #.setParseAction(getSign)


class SimpleNumber(object):
    def __repr__(self):
        return "Number(%s)" % (self.text,)

    def __str__(self):
        return self.__repr__()

    def __init__(self, text):
        self.text = text
        self.v = float(text)



def _simple_number():
    s = Combine(optional_sign + usn)
    s.setParseAction(lambda s, l, tok: SimpleNumber(tok[0]))

    return s

simple_number = _simple_number()

def test_usn():
    for f in ["32.4", "0.23", "0.3e-7", "1.234e+7"]:
        assert usn.parseString(f)[0] == f
        #assert usn.parseString(f)[0][0] == float(f)



class SimpleInteger(object):
    def __repr__(self):
        return "Number(%s)" % (self.text,)

    def __str__(self):
        return self.__repr__()

    def __init__(self, text):
        self.text = text
        self.v = int(text)

def _unsigned_integer():
    s = Combine(Optional("+") + Word(nums))
    s.setParseAction(lambda s, l, tok: SimpleInteger(tok[0]))

    return s


simple_integer = _unsigned_integer()

def test_integer():
    for f in ["32", "+3"]:
        assert simple_integer.parseString(f)[0].text == f


def get_degree(s, l, tok):
    pass


class Sixty(object):
    def __init__(self, sn, d, m, s):
        self.v = sn * (d +(m + s/60.)/60.)
        self.degree = self.v

class HMS(object):
    def __repr__(self):
        return "HMS(%s)" % (self.text,)

    def __init__(self, kl):
        self.text = "".join(kl)

        if kl[0] == "-":
            sn = -1
        else:
            sn = +1

        kkl = kl[1::2]
        if len(kkl) == 3:
            d, m, s = float(kl[1]), float(kl[3]), float(kl[5])
        elif len(kkl) == 2:
            d, m, s = float(kl[1]), float(kl[3]), 0.
        else:
            d, m, s = float(kl[1]), 0., 0.


        self.v = sn * (d +(m + s/60.)/60.)
        self.degree = self.v * 15


class DMS(object):
    def __repr__(self):
        return "DMS(%s)" % (self.text,)

    def __init__(self, kl):
        self.text = "".join(kl)

        if kl[0] == "-":
            sn = -1
        else:
            sn = +1

        kkl = kl[1::2]
        if len(kkl) == 3:
            d, m, s = float(kl[1]), float(kl[3]), float(kl[5])
        elif len(kkl) == 2:
            d, m, s = float(kl[1]), float(kl[3]), 0.
        else:
            d, m, s = float(kl[1]), 0., 0.


        self.v = sn * (d +(m + s/60.)/60.)
        self.degree = self.v



class AngularDistance(object):
    def __repr__(self):
        return "Ang(%s)" % (self.text,)

    def __init__(self, kl):
        self.text = "".join(kl)

        m, s = 0, 0
        if kl[1] == "'":
            m = float(kl[0])
            if len(kl) == 4:
                s = float(kl[2])
        else:
            s = float(kl[0])

        self.v = (m + s/60.)/60.
        self.degree = self.v


def _sexadecimal():
    colon = Literal(":")

    s = optional_sign + usn + colon + usn + \
        Optional(colon + usn)

        #Optional(colon + usn).setParseAction(get_default(0))

    #s = s.leaveWhitespace()

    return s

#sexadecimal = _sexadecimal()

sexadecimal60 = _sexadecimal().setParseAction(lambda s, l, tok: DMS(tok))
sexadecimal24 = _sexadecimal().setParseAction(lambda s, l, tok: HMS(tok))


def test_sexadecimal():
    s = sexadecimal60.parseString

    assert s("32:24:32.2")[0].v == Sixty(1, 32, 24, 32.2).v
    assert s("-32:24:32.2")[0].v == Sixty(-1, 32, 24, 32.2).v
    assert s("+32:24:32.2")[0].v == Sixty(1, 32, 24, 32.2).v



def _hms_number():
    _h = (usn + Literal("h")).leaveWhitespace()
    _m = (usn + Literal("m")).leaveWhitespace()
    _s = (usn + Literal("s")).leaveWhitespace()


    hms = optional_sign + _h + Optional(_m + Optional(_s))

    hms = hms.setParseAction(lambda s, l, tok: HMS(tok))

    return hms


def _dms_number():
    _d = (usn + Literal("d")).leaveWhitespace()
    _m = (usn + Literal("m")).leaveWhitespace()
    _s = (usn + Literal("s")).leaveWhitespace()


    dms = optional_sign + _d + Optional(_m + Optional(_s))

    dms = dms.setParseAction(lambda s, l, tok: DMS(tok))

    return dms


    #zzz = Empty().setParseAction(lambda s, l, tok: 0)


    #_hms = optional_sign + Or([_h + Optional(_m).setParseAction(get_default(0)) \
    #                           + Optional(_s).setParseAction(get_default(0)),
    #                           zzz + _m + Optional(_s).setParseAction(get_default(0)),
    #                           zzz + zzz + _s])

    #_hms = _hms.leaveWhitespace()



hms_number = _hms_number()
dms_number = _dms_number()


def test_hms():
    s = hms_number.parseString

    assert s("32h24m32.2s")[0].v == Sixty(1, 32, 24, 32.2).v
    assert s("0h24m32.2s")[0].v == Sixty(1, 0, 24, 32.2).v
    assert s("32h")[0].v == Sixty(1, 32, 0, 0).v



# def _dms_number():
#     _d = (usn + Literal("d")).leaveWhitespace()
#     _m = (usn + Literal("m")).leaveWhitespace()
#     _s = (usn + Literal("s")).leaveWhitespace()

#     zzz = Empty().setParseAction(lambda s, l, tok: 0)

#     _dms = optional_sign + Or([_d + Optional(_m).setParseAction(get_default(0)) + Optional(_s).setParseAction(get_default(0)),
#                                zzz + _m + Optional(_s).setParseAction(get_default(0)),
#                                zzz + zzz + _s])

#     #_hms = _hms.leaveWhitespace()
#     dms = _dms.setParseAction(lambda s, l, tok: DMS(*tok))

#     return dms



def test_dms():
    s = dms_number.parseString

    assert s("32d24m32.2s")[0].v == Sixty(1, 32, 24, 32.2).v
    assert s("-32d24m32.2s")[0].v == Sixty(-1, 32, 24, 32.2).v
    assert s("32d")[0].v == Sixty(1, 32, 0, 0).v



def _angular_distance():
    _m = (usn + Literal("\'")).leaveWhitespace()
    _s = (usn + Literal("\"")).leaveWhitespace()

    ms = Or([ _m + Optional(_s),
               _s])

    #_hms = _hms.leaveWhitespace()
    ms = ms.setParseAction(lambda s, l, tok: AngularDistance(tok))

    return ms


angular_distance = _angular_distance()

def test_ang_distance():
    s = angular_distance.parseString

    assert s("32.3'")[0].v == Sixty(1, 0, 32.3, 0.).v
    assert s("32\'24\"")[0].v == Sixty(1, 0, 32, 24).v





class Arg(object):
    def __init__(self, type, parser):
        self.type = type
        self.parser = parser


class CoordOdd:
    parser = (hms_number | sexadecimal24 | simple_number)
    type = HMS

class CoordEven:
    parser = (dms_number | sexadecimal60 | simple_number)
    type = DMS

class Distance:
    parser = (angular_distance | simple_number)
    type = AngularDistance

class Angle:
    parser = (simple_number)
    type = simple_number

class Integer:
    parser = simple_integer
    type = simple_number


# CoordOdd = Arg(type=HMS,
#                parser=(hms_number | sexadecimal24 | simple_number)
#                )

# CoordEven = Arg(type=DMS,
#                 parser=(dms_number | sexadecimal60 | simple_number)
#                 )

# Distance = Arg(type=AngularDistance,
#                parser=(angular_distance | simple_number)
#                )

# Angle = Arg(type=float,
#             parser=(simple_number)
#             )



def test_coord_odd():
    s = CoordOdd.parser.parseString

    assert s("32h24m32.2s")[0].v == Sixty(1, 32, 24, 32.2).v
    assert s("32:24:32.2s")[0].v == Sixty(1, 32, 24, 32.2).v
    assert s("32.24")[0].v == 32.24

    s1 = s("32:24:32.2s")[0]
    assert isinstance(s1, HMS)


def test_coord_odd():
    s = CoordOdd.parser.parseString

    assert s("32h24m32.2s")[0].v == Sixty(1, 32, 24, 32.2).v
    assert s("32:24:32.2s")[0].v == Sixty(1, 32, 24, 32.2).v
    assert s("32.24")[0].v == 32.24

    s1 = s("32:24:32.2s")[0]
    assert isinstance(s1, HMS)



import sys
import warnings

from setuptools import setup, Extension

# check if cython or pyrex is available.
pyrex_impls = 'Cython.Distutils.build_ext', 'Pyrex.Distutils.build_ext'
for pyrex_impl in pyrex_impls:
    try:
        # from (pyrex_impl) import build_ext
        build_ext = __import__(pyrex_impl, fromlist=['build_ext']).build_ext
        break
    except ImportError:
        pass
have_pyrex = 'build_ext' in globals()

if have_pyrex:
    cmdclass = {'build_ext': build_ext}
    PYREX_SOURCE = "src/_region_filter.pyx"
else:
    cmdclass = {}
    PYREX_SOURCE = "src/_region_filter.c"

# If you don't want to build filtering module (which requires a C compiler), set it to False
WITH_FILTER = True

arguments = dict()

if WITH_FILTER:
    try:
        import numpy
    except ImportError:
        warnings.warn(
            "numpy must be installed to build the filtering module.")
        sys.exit(1)

    try:
        numpy_include = numpy.get_include()
    except AttributeError:
        numpy_include = numpy.get_numpy_include()

    if len(cmdclass) > 0:
        arguments["cmdclass"] = cmdclass

    arguments["ext_modules"] = [
        Extension(
            "stregion._region_filter",
            [PYREX_SOURCE],
            include_dirs=[
                './src',
                numpy_include,
            ],
            libraries=[],
        )
    ]

setup(**arguments)



import pytest

import stregion
import os

basedir = os.path.dirname(stregion.__file__)

lib_files_with_test = ["ds9_region_parser.py",
                       "region_numbers.py",
                       "ds9_attr_parser.py",
                       "parser_helper.py"]

pytest.main([os.path.join(basedir, f) for f in lib_files_with_test] \
            + ['tests'])



import copy

from pyparsing import Literal, CaselessKeyword, CaselessLiteral, \
     Word, Optional, OneOrMore, Group, Combine, ZeroOrMore, nums, \
     Forward, StringEnd, restOfLine, alphas, alphanums, CharsNotIn, \
     MatchFirst, And, Or, quotedString, QuotedString, White

from .region_numbers import CoordOdd, CoordEven, Distance, Angle

from .parser_helper import as_comma_separated_list, wcs_shape, \
     define_shape, define_shape_helper, define_expr, define_line, \
     comment_shell_like, define_simple_literals, \
     Shape

def get_ds9_attr_parser():
    lhs = Word(alphas)
    paren = QuotedString("(",endQuoteChar=")")
    rhs = Or([Word(alphas+nums),
              Combine(Word(alphas) + White() + Word(nums)), # for point
              quotedString,
              QuotedString("{",endQuoteChar="}"),
              paren + ZeroOrMore(paren),
              Word(nums+" "),
              Word(nums+".")
              ])
    expr = lhs + Optional(Literal("=").suppress() + rhs)
    expr.setParseAction(lambda s, l, tok: tuple(tok))

    return ZeroOrMore(expr)



class Ds9AttrParser(object):
    def set_continued(self, s, l, tok):
        self.continued = True
        
    def __init__(self):
        self.continued = False

        ds9_attr_parser = get_ds9_attr_parser()

        ds9_shape_in_comment_defs = dict(text=wcs_shape(CoordOdd, CoordEven),
                                         vector=wcs_shape(CoordOdd, CoordEven,
                                                          Distance, Angle),
                                         composite=wcs_shape(CoordOdd, CoordEven, Angle),
                                         )
        regionShape = define_shape_helper(ds9_shape_in_comment_defs)
        regionShape = regionShape.setParseAction(lambda s, l, tok: Shape(tok[0], tok[1:]))


        self.parser_default = ds9_attr_parser

        cont = CaselessKeyword("||").setParseAction(self.set_continued).suppress()
        line = Optional(And([regionShape,
                             Optional(cont)])) \
                             + ds9_attr_parser

        self.parser_with_shape = line


    def parse_default(self, s):
        return self.parser_default.parseString(s)

    def parse_check_shape(self, s):
        l = self.parser_with_shape.parseString(s)
        if l and isinstance(l[0], Shape):
            if self.continued:
                l[0].continued = True
            return l[0], l[1:]
        else:
            return None, l



def get_attr(attr_list, global_attrs):
    local_attr = [], {}
    for kv in attr_list:
        if len(kv) == 1:
            local_attr[0].append(kv[0])
        elif len(kv) == 2:
            local_attr[1][kv[0]] = kv[1]
        elif len(kv) > 2:
            local_attr[1][kv[0]] = kv[1:]

    attr0 = copy.copy(global_attrs[0])
    attr1 = copy.copy(global_attrs[1])

    if local_attr[0]:
        attr0.extend(local_attr[0])

    if local_attr[1]:
        attr1.update(local_attr[1])

    return attr0, attr1






def test_attr():

    p = get_ds9_attr_parser()
    assert  p.parseString("color = green")[0] == ("color", "green")
    assert  p.parseString("font=\"123 123\"")[0] == ("font", '"123 123"')
    assert  p.parseString("color")[0] == ("color",)
    assert  p.parseString("tag={group 1}")[0] == ("tag","group 1")

if __name__ == "__main__":
    p = get_ds9_attr_parser()
    s = 'point=cross 40 color=green dashlist= 8 4 font="helvetica 10 normal" tag={group 1} select=1 source panda=(1 3 2)(2 3 4)'
    #s = 'dashlist= 8 4 font=1'
    ss = p.parseString(s)



from pyparsing import Literal, CaselessKeyword, CaselessLiteral, \
     Word, Optional, OneOrMore, Group, Combine, ZeroOrMore, nums, \
     Forward, StringEnd, restOfLine, alphas, alphanums, CharsNotIn, \
     MatchFirst, And, Or


def as_comma_separated_list(al):

    l = [al[0]]
    comma = Literal(",").suppress()

    for a1 in al[1:]:
        l.append(comma)
        l.append(a1)

    return And(l)


class wcs_shape(object):
    def __init__(self, *kl, **kw):
        self.args_list = kl
        self.args_repeat = kw.get("repeat", None)

    def get_pyparsing(self):
        return [a.parser for a in self.args_list]


def define_shape(name, shape_args, args_repeat=None):
    lparen = Literal("(").suppress()
    rparen = Literal(")").suppress()
    comma = Literal(",").suppress()

    shape_name = CaselessKeyword(name)


    if args_repeat is None:
        shape_with_parens = And([shape_name, lparen,
                                 as_comma_separated_list(shape_args),
                                 rparen])

        shape_with_spaces = shape_name + And(shape_args)

    else:
        n1, n2 = args_repeat
        sl = []

        ss = shape_args[:n1]
        if ss:
            sl.append(as_comma_separated_list(ss))

        ss = shape_args[n1:n2]
        if ss:
            ar = as_comma_separated_list(ss)
            if sl:
                sl.extend([comma+ ar, ZeroOrMore(comma + ar)])
            else:
                sl.extend([ar, ZeroOrMore(comma + ar)])

        ss = shape_args[n2:]
        if ss:
            if sl:
                sl.extend([comma, as_comma_separated_list(ss)])
            else:
                sl.extend([as_comma_separated_list(ss)])

        sl = [shape_name, lparen] + sl + [rparen]

        shape_with_parens = And(sl)

        shape_with_spaces = shape_name + OneOrMore(And(shape_args))


    return (shape_with_parens | shape_with_spaces)


def test_define_shape():
    from .region_numbers import CoordOdd, CoordEven, Distance
    args = [s.parser for s in [CoordOdd, CoordEven, Distance]]
    circle_parser = define_shape("circle", args, args_repeat=None)

    p = circle_parser.parseString("circle(1:2:3, 2:4:5, 3.)")
    assert p[0] == "circle"
    assert isinstance(p[1], CoordOdd.type)
    assert isinstance(p[2], CoordEven.type)

    args = [s.parser for s in [CoordEven]]
    circle_parser = define_shape("circle", args, args_repeat=None)

    p = circle_parser.parseString("circle(1:2:3)")
    p = circle_parser.parseString("circle(1d2m3s)")
    assert p[0] == "circle"
    assert isinstance(p[1], CoordEven.type)

    args = [s.parser for s in [CoordOdd, CoordEven, Distance, Distance]]
    ell_parser = define_shape("ell", args, args_repeat=(2,4))

    p = ell_parser.parseString("ell(1:2:2, 2:2:2, 3, 4, 5, 6)")
    assert p[0] == "ell"
    assert isinstance(p[1], CoordOdd.type)
    assert isinstance(p[2], CoordEven.type)

    args = [s.parser for s in [CoordOdd, CoordEven]]
    polygon_parser = define_shape("polygon", args, args_repeat=(0,2))
    p = polygon_parser.parseString("polygon(3:2:4.22, 3:3:4., 3:2:3, 3.2)")
    assert p[0] == "polygon"
    assert isinstance(p[-2], CoordOdd.type)
    assert isinstance(p[2], CoordEven.type)
    #assert isinstance(p[-1], float)


def define_shape_helper(shape_defs):
    l = []

    for n, args in shape_defs.items():
        s = define_shape(n,
                         args.get_pyparsing(),
                         args_repeat = args.args_repeat)

        l.append(s)

    return Or(l)



def define_expr(regionShape, negate_func):

    minus = Literal("-").suppress()
    regionExclude = (minus + regionShape).setParseAction(negate_func)
    regionExpr = (regionShape | regionExclude)

    return regionExpr


def define_line(atom,
                separator,
                comment):

    atomSeparator = separator.suppress()

    atomList = atom + ZeroOrMore(atomSeparator + atom)

    line = (atomList + Optional(comment)) | \
           comment

    return line




def comment_shell_like(comment_begin, parseAction=None):

    c = comment_begin + restOfLine
    if parseAction:
        c = c.setParseAction(parseAction)

    return c


def define_simple_literals(literal_list, parseAction=None):

    l = MatchFirst([CaselessKeyword(k) for k in literal_list])

    if parseAction:
        l = l.setParseAction(parseAction)

    return l


class Shape(object):

    def __init__(self, shape_name, shape_params):
        self.name = shape_name
        self.params = shape_params

        self.comment = None
        self.exclude = False
        self.continued = False
        
    def __repr__(self):
        params_string = ",".join(map(repr, self.params))
        if self.exclude:
            return "Shape : -%s ( %s )" %  (self.name, params_string)
        else:
            return "Shape : %s ( %s )" %  (self.name, params_string)

    def set_exclude(self):
        self.exclude = True


class Property(object):
    def __init__(self, text):
        self.text = text

    def __repr__(self):
        return "Property : " + repr(self.text)

class CoordCommand(Property):
    def __repr__(self):
        return "CoordCommand : " + repr(self.text)

class Global(Property):
    def __repr__(self):
        return "Global : " + repr(self.text)

class Comment(Property):
    def __repr__(self):
        return "Comment : " + repr(self.text)


class RegionPusher(object):
    def __init__(self):
        self.flush()

    def flush(self):
        self.stack = []
        self.comment = None
        self.continued = None
        
    def pushAtom(self, s, l, tok):
        self.stack.append(tok[-1])

    def pushComment(self, s, l, tok):
        self.comment  = tok[-1].strip()

    def set_continued(self, s, l, tok):
        self.continued = True

